"""Log output fixtures.

To capture log output during tests, you can use `log_output` fixture.
If you need to add custom processors or define `log_level`, please use `log_level` and `log_processors` fixtures.

Example:
    @pytest.mark.usefixtures('configure_structlog')
    def test_log_output(log_ouput):
        assert log_output.entries == []
        # do something
        assert log_output.entries == [...]

"""
import logging

import pytest
import structlog
from outcome.logkit.init import get_final_processors
from structlog.testing import LogCapture


@pytest.fixture
def log_output():  # pragma: no cover
    return LogCapture()


@pytest.fixture
def log_level():  # pragma: no cover
    return logging.DEBUG


@pytest.fixture
def log_processors(log_output):  # pragma: no cover
    return [log_output]


@pytest.fixture
def configure_structlog(log_level, log_processors):  # pragma: no cover
    processors = get_final_processors(log_level, log_processors)
    structlog.configure(processors=processors)



"""Interface to structlog."""

from structlog.contextvars import bind_contextvars, clear_contextvars, unbind_contextvars

# For these to work, structlog needs to be configured with the structlog.contextvars.merge_contextvars processor
# This is handled for you when you use `init()`


def add(**kwargs):
    bind_contextvars(**kwargs)


def remove(*args):
    unbind_contextvars(*args)


def clear():
    clear_contextvars()



"""Logging proxy. Logs calls to attributes and methods."""

import logging
from typing import Any

from outcome.logkit.logger import get_logger


class LoggingProxy:
    def __init__(self, target, level=logging.DEBUG, name=None):
        self._target = target
        self._logger = get_logger(name)
        self._name = name
        self._level = level

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        try:  # noqa: WPS501
            rv = self._target(*args, **kwargs)
        finally:
            self._logger.log(
                self._name, type='method', args=args, kwargs=kwargs, retval=rv, levelno=self._level, logger=self._name,
            )

        return rv

    def __getattr__(self, name):
        attr = getattr(self._target, name)

        if callable(attr):
            return LoggingProxy(attr, name=f'{self._name}.{name}', level=self._level)
        else:
            self._logger.log(
                f'{self._name}.{name}', type='attribute', retval=attr, levelno=self._level, logger=f'{self._name}.{name}',
            )

        return attr



"""Helper filter for logging."""

import logging


# Excludes records that match the name provided to the
# filter on init
class ExcludeFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:  # noqa: A003,WPS125
        return not super().filter(record)
